@@ -2,7 +2,7 @@
  * VLCAppCoordinator.m
  * VLC for iOS
  *****************************************************************************
- * Copyright (c) 2022 VideoLAN. All rights reserved.
+ * Copyright (c) 2022-2024 VideoLAN. All rights reserved.
  * $Id$
  *
  * Author: Felix Paul Kühne <fkuehne # videolan.org>
@@ -13,12 +13,14 @@
 #import "VLCAppCoordinator.h"
 #import <CoreSpotlight/CoreSpotlight.h>
 #import "VLCRemoteControlService.h"
+#import "VLCFavoriteService.h"
 #import "VLC-Swift.h"
 
 @interface VLCAppCoordinator()
 {
     MediaLibraryService *_mediaLibraryService;
     VLCRendererDiscovererManager *_rendererDiscovererManager;
+    VLCFavoriteService *_favoriteService;
     VLCHTTPUploaderController *_httpUploaderController;
     UITabBarController *_tabBarController;
     TabBarCoordinator *_tabCoordinator;
@@ -47,24 +49,41 @@
 {
     self = [super init];
     if (self) {
-        _mediaLibraryService = [[MediaLibraryService alloc] init];
-
-        // Init the HTTP Server and clean its cache
-        // FIXME: VLCHTTPUploaderController should perhaps be a service?
-        _httpUploaderController = [VLCHTTPUploaderController sharedInstance];
-        [_httpUploaderController cleanCache];
-        _httpUploaderController.medialibrary = _mediaLibraryService;
-
-        _remoteControlService = [[VLCRemoteControlService alloc] init];
+        dispatch_async(dispatch_get_main_queue(), ^{
+            [self initializeServices];
+        });
     }
     return self;
 }
 
+- (void)initializeServices
+{
+    // Init the HTTP Server and clean its cache
+    _httpUploaderController = [[VLCHTTPUploaderController alloc] init];
+    [_httpUploaderController cleanCache];
+    _httpUploaderController.medialibrary = self.mediaLibraryService;
+
+    // start the remote control service
+    _remoteControlService = [[VLCRemoteControlService alloc] init];
+}
+
 - (MediaLibraryService *)mediaLibraryService
 {
+    if (!_mediaLibraryService) {
+        _mediaLibraryService = [[MediaLibraryService alloc] init];
+    }
     return _mediaLibraryService;
 }
 
+- (VLCFavoriteService *)favoriteService
+{
+    if (!_favoriteService) {
+        _favoriteService = [[VLCFavoriteService alloc] init];
+    }
+
+    return _favoriteService;
+}
+
 - (VLCRendererDiscovererManager *)rendererDiscovererManager
 {
     if (!_rendererDiscovererManager) {
@@ -74,6 +93,14 @@
     return _rendererDiscovererManager;
 }
 
+- (VLCHTTPUploaderController *)httpUploaderController
+{
+    if (!_httpUploaderController) {
+        [self initializeServices];
+    }
+    return _httpUploaderController;
+}
+
 - (void)setExternalWindow:(UIWindow *)externalWindow
 {
     _externalWindow = externalWindow;
@@ -105,7 +132,6 @@
     _tabCoordinator = [[TabBarCoordinator alloc] initWithTabBarController:_tabBarController mediaLibraryService:self.mediaLibraryService];
 
     _playerDisplayController = [[VLCPlayerDisplayController alloc] init];
-    [_tabBarController addChildViewController:_playerDisplayController];
     [_tabBarController.view addSubview:_playerDisplayController.view];
     _playerDisplayController.view.layoutMargins = UIEdgeInsetsMake(0, 0, tabBarController.tabBar.frame.size.height, 0);
     _playerDisplayController.realBottomAnchor = tabBarController.tabBar.topAnchor;
@@ -129,7 +155,7 @@
     }
 
     if (identifier > 0) {
-        return [_mediaLibraryService mediaFor:identifier];
+        return [self.mediaLibraryService mediaFor:identifier];
     }
 
     return nil;
