@@ -25,6 +25,10 @@
 #import "VLCPlayerDisplayController.h"
 #import <stdatomic.h>
 
+#if TARGET_OS_IOS
+#import "VLCMLMedia+Podcast.h"
+#endif
+
 #import "VLC-Swift.h"
 
 NSString *const VLCPlaybackServicePlaybackDidStart = @"VLCPlaybackServicePlaybackDidStart";
@@ -54,13 +58,12 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
     NSTimer *_sleepTimer;
 
     BOOL _isInFillToScreen;
-    NSUInteger _previousAspectRatio;
+    NSInteger _previousAspectRatio;
 
     UIView *_videoOutputViewWrapper;
     UIView *_actualVideoOutputView;
     UIView *_preBackgroundWrapperView;
 
-    BOOL _mediaWasJustStarted;
     int _majorPositionChangeInProgress;
     BOOL _externalAudioPlaybackDeviceConnected;
 
@@ -150,6 +153,11 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 
 #pragma mark - playback management
 
+- (void)addAudioToCurrentPlaybackFromURL:(NSURL *)audioURL
+{
+    [_mediaPlayer addPlaybackSlave:audioURL type:VLCMediaPlaybackSlaveTypeAudio enforce:YES];
+}
+
 - (void)addSubtitlesToCurrentPlaybackFromURL:(NSURL *)subtitleURL
 {
     [_mediaPlayer addPlaybackSlave:subtitleURL type:VLCMediaPlaybackSlaveTypeSubtitle enforce:YES];
@@ -289,18 +297,11 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
     int deinterlace = [[defaults objectForKey:kVLCSettingDeinterlace] intValue];
     [_mediaPlayer setDeinterlace:deinterlace withFilter:@"blend"];
 
-    VLCMedia *media = [_mediaList mediaAtIndex:_itemInMediaListToBePlayedFirst];
-    [media parseWithOptions:VLCMediaParseLocal];
-    media.delegate = self;
-    [media addOptions:self.mediaOptionsDictionary];
-
     [_listPlayer setMediaList:self.mediaList];
-#if TARGET_OS_IOS
     if ([defaults boolForKey:kVLCPlayerShouldRememberState]) {
         VLCRepeatMode repeatMode = [defaults integerForKey:kVLCPlayerIsRepeatEnabled];
         [_listPlayer setRepeatMode:repeatMode];
     }
-#endif
 
     [_playbackSessionManagementLock unlock];
 
@@ -334,21 +335,36 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
         }
     }
 
-    _mediaWasJustStarted = YES;
-
     [_mediaPlayer addObserver:self forKeyPath:@"time" options:0 context:nil];
-    [_mediaPlayer addObserver:self forKeyPath:@"remainingTime" options:0 context:nil];
 
 #if TARGET_OS_IOS
     [_mediaPlayer setRendererItem:_renderer];
 #endif
 
+    /* we are playing a collection without a valid index,
+     * so this is either 0 or totally random */
+    if (_itemInMediaListToBePlayedFirst == -1) {
+        if (_shuffleMode) {
+            int count = (int)_mediaList.count;
+            if (count > 0) {
+                _currentIndex = arc4random_uniform(count - 1);
+                [self shuffleMediaList];
+                _itemInMediaListToBePlayedFirst = (int)[_shuffledOrder[0] integerValue];
+            }
+        } else {
+            _itemInMediaListToBePlayedFirst = 0;
+        }
+    }
+
+    VLCMedia *media = [_mediaList mediaAtIndex:_itemInMediaListToBePlayedFirst];
+    [media parseWithOptions:VLCMediaParseLocal];
+    media.delegate = self;
+    [media addOptions:self.mediaOptionsDictionary];
+
     [_listPlayer playItemAtNumber:@(_itemInMediaListToBePlayedFirst)];
 
     _currentIndex = _itemInMediaListToBePlayedFirst;
-    if(_shuffleMode) {
-        [self shuffleMediaList];
-    }
+
     if ([self.delegate respondsToSelector:@selector(prepareForMediaPlayback:)])
         [self.delegate prepareForMediaPlayback:self];
 
@@ -371,7 +387,6 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 
     _playerIsSetup = YES;
 
-    [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackDidStart object:self];
     [_playbackSessionManagementLock unlock];
 }
 
@@ -387,7 +402,6 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
     if (_mediaPlayer) {
         @try {
             [_mediaPlayer removeObserver:self forKeyPath:@"time"];
-            [_mediaPlayer removeObserver:self forKeyPath:@"remainingTime"];
         }
         @catch (NSException *exception) {
             APLog(@"we weren't an observer yet");
@@ -454,22 +468,6 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 
 - (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context
 {
-    if (_mediaWasJustStarted) {
-        _mediaWasJustStarted = NO;
-#if TARGET_OS_IOS
-        if (self.mediaList) {
-            [self _recoverLastPlaybackState];
-        }
-#else
-        NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
-        BOOL bValue = [defaults boolForKey:kVLCSettingUseSPDIF];
-
-        if (bValue) {
-           _mediaPlayer.audio.passthrough = bValue;
-        }
-#endif
-    }
-
     if ([self.delegate respondsToSelector:@selector(playbackPositionUpdated:)])
         [self.delegate playbackPositionUpdated:self];
 
@@ -511,6 +509,11 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
         [self.delegate playModeUpdated];
     }
     [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackModeUpdated object:self];
+
+    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
+    if ([defaults boolForKey:kVLCPlayerShouldRememberState]) {
+        [defaults setInteger:repeatMode forKey:kVLCPlayerIsRepeatEnabled];
+    }
 }
 
 - (BOOL)currentMediaHasChapters
@@ -532,12 +535,11 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 
 - (BOOL)isNextMediaAvailable
 {
-    NSLog(@"%s", __func__);
     if (_mediaList.count == 1) {
         return NO;
     }
 
-    if ([self nextMediaIndex:NO] > 0) {
+    if (_currentIndex < _mediaList.count - 1) {
         return YES;
     }
 
@@ -593,16 +595,14 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 
 - (void)toggleRepeatMode
 {
+    VLCRepeatMode newRepeatMode;
     if (_listPlayer.repeatMode == VLCRepeatAllItems) {
-        _listPlayer.repeatMode = VLCDoNotRepeat;
+        newRepeatMode = VLCDoNotRepeat;
     } else {
-        _listPlayer.repeatMode += 1;
+        newRepeatMode = _listPlayer.repeatMode + 1;
     }
 
-    if ([self.delegate respondsToSelector:@selector(playModeUpdated)]) {
-        [self.delegate playModeUpdated];
-    }
-    [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackModeUpdated object:self];
+    [self setRepeatMode:newRepeatMode];
 }
 
 - (NSInteger)indexOfCurrentAudioTrack
@@ -625,14 +625,19 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
     return _mediaPlayer.currentTitleIndex;
 }
 
+- (NSInteger)numberOfVideoTracks
+{
+    return [_mediaPlayer numberOfVideoTracks];
+}
+
 - (NSInteger)numberOfAudioTracks
 {
-    return [_mediaPlayer numberOfAudioTracks];
+    return [_mediaPlayer numberOfAudioTracks] + 1;
 }
 
 - (NSInteger)numberOfVideoSubtitlesIndexes
 {
-    return _mediaPlayer.videoSubTitlesIndexes.count + 1;
+    return _mediaPlayer.videoSubTitlesIndexes.count + 2;
 }
 
 - (NSInteger)numberOfTitles
@@ -651,6 +656,9 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
     if (index >= 0 && index < count) {
         return _mediaPlayer.videoSubTitlesNames[index];
     } else if (index == count) {
+        return NSLocalizedString(@"SELECT_SUBTITLE_FROM_FILES", nil);
+    }
+    else if (index == count + 1) {
         return NSLocalizedString(@"DOWNLOAD_SUBS_FROM_OSO", nil);
     }
     return @"";
@@ -660,6 +668,8 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 {
     if (index >= 0 && index < _mediaPlayer.audioTrackNames.count)
         return _mediaPlayer.audioTrackNames[index];
+    else if (index == _mediaPlayer.audioTrackNames.count)
+        return NSLocalizedString(@"SELECT_AUDIO_FROM_FILES", nil);
     return @"";
 }
 
@@ -749,6 +759,31 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 #pragma clang diagnostic pop
         } break;
 
+        case VLCMediaPlayerStateOpening: {
+#if TARGET_OS_IOS
+            [self _recoverLastPlaybackState];
+#else
+            NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
+            BOOL bValue = [defaults boolForKey:kVLCSettingUseSPDIF];
+
+            if (bValue) {
+                _mediaPlayer.audio.passthrough = bValue;
+            }
+#endif
+            [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackDidStart object:self];
+        } break;
+
+        case VLCMediaPlayerStatePlaying: {
+            [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackDidResume object:self];
+        } break;
+
+        case VLCMediaPlayerStatePaused: {
+#if TARGET_OS_IOS
+            [self savePlaybackState];
+#endif
+            [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackDidPause object:self];
+        } break;
+
         case VLCMediaPlayerStateError: {
             APLog(@"Playback failed");
             dispatch_async(dispatch_get_main_queue(),^{
@@ -786,6 +821,8 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
             break;
     }
 
+    _mediaPlayerState = currentState;
+
     if ([self.delegate respondsToSelector:@selector(mediaPlayerStateChanged:isPlaying:currentMediaHasTrackToChooseFrom:currentMediaHasChapters:forPlaybackService:)])
         [self.delegate mediaPlayerStateChanged:currentState
                                      isPlaying:_mediaPlayer.isPlaying
@@ -793,12 +830,9 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
                        currentMediaHasChapters:self.currentMediaHasChapters
                          forPlaybackService:self];
 
-    [self setNeedsMetadataUpdate];
-}
-
-- (void)setPlayAsAudio:(BOOL)playAsAudio
-{
-    _playAsAudio = playAsAudio;
+    dispatch_async(dispatch_get_main_queue(), ^{
+        [self setNeedsMetadataUpdate];
+    });
 }
 
 #pragma mark - playback controls
@@ -810,16 +844,11 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 - (void)play
 {
     [_listPlayer play];
-    [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackDidResume object:self];
 }
 
 - (void)pause
 {
     [_listPlayer pause];
-#if TARGET_OS_IOS
-    [self savePlaybackState];
-#endif
-    [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackDidPause object:self];
 }
 
 - (void)playItemAtIndex:(NSUInteger)index
@@ -847,6 +876,11 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
         [self.delegate playModeUpdated];
     }
     [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackServicePlaybackModeUpdated object:self];
+
+    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
+    if ([[defaults valueForKey:kVLCPlayerShouldRememberState] boolValue]) {
+        [defaults setBool:shuffleMode forKey:kVLCPlayerIsShuffleEnabled];
+    }
 }
 
 - (void)shuffleMediaList {
@@ -900,12 +934,12 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
         }
     }
 
-    if (_shuffleMode && mediaListCount > 2) {
-        nextIndex = [_shuffledOrder[_currentIndex] integerValue];
-    }
-
     _currentIndex = nextIndex;
 
+    if (_shuffleMode && mediaListCount > 2 && _currentIndex >= 0 && _currentIndex < _shuffledOrder.count) {
+        return [_shuffledOrder[_currentIndex] integerValue];
+    }
+
     return _currentIndex;
 }
 
@@ -1030,12 +1064,27 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
         _currentAspectRatio = _isInFillToScreen ? _previousAspectRatio : VLCAspectRatioFillToScreen;
     } else {
         // Increment unless hitting last aspectratio
-        _currentAspectRatio = _currentAspectRatio == VLCAspectRatioSixteenToTen ? VLCAspectRatioDefault : _currentAspectRatio + 1;
+        _currentAspectRatio = _currentAspectRatio == VLCAspectRatioThirtyNineToOne ? VLCAspectRatioDefault : _currentAspectRatio + 1;
     }
 
     // If fullScreen is toggled directly and then the aspect ratio changes, fullScreen is not reset
     if (_isInFillToScreen) _isInFillToScreen = NO;
 
+    [self applyAspectRatio];
+
+    if ([self.delegate respondsToSelector:@selector(showStatusMessage:)]) {
+        [self.delegate showStatusMessage:[NSString stringWithFormat:NSLocalizedString(@"AR_CHANGED", nil), [VLCAspectRatioBridge stringToDisplayFor:_currentAspectRatio]]];
+    }
+}
+
+- (void)setCurrentAspectRatio:(NSInteger)currentAspectRatio
+{
+    _currentAspectRatio = currentAspectRatio;
+    [self applyAspectRatio];
+}
+
+- (void)applyAspectRatio
+{
     switch (_currentAspectRatio) {
         case VLCAspectRatioDefault:
             _mediaPlayer.scaleFactor = 0;
@@ -1051,44 +1100,27 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
             [self switchToFillToScreen];
             break;
         case VLCAspectRatioFourToThree:
+        case VLCAspectRatioFiveToFour:
         case VLCAspectRatioSixteenToTen:
         case VLCAspectRatioSixteenToNine:
+        case VLCAspectRatioTwentyOneToOne:
+        case VLCAspectRatioThirtyFiveToOne:
+        case VLCAspectRatioThirtyNineToOne:
             _mediaPlayer.scaleFactor = 0;
+            NSString *aspectRatio = [VLCAspectRatioBridge valueFor:_currentAspectRatio];
 #if LIBVLC_VERSION_MAJOR == 3
             _mediaPlayer.videoCropGeometry = NULL;
-            _mediaPlayer.videoAspectRatio = (char *)[[self stringForAspectRatio:_currentAspectRatio] UTF8String];
+            _mediaPlayer.videoAspectRatio = (char *)[aspectRatio UTF8String];
 #else
-            _mediaPlayer.videoAspectRatio = [self stringForAspectRatio:_currentAspectRatio];
+            _mediaPlayer.videoAspectRatio = aspectRatio;
 #endif
     }
 
-    if ([self.delegate respondsToSelector:@selector(showStatusMessage:)]) {
-        [self.delegate showStatusMessage:[NSString stringWithFormat:NSLocalizedString(@"AR_CHANGED", nil), [self stringForAspectRatio:_currentAspectRatio]]];
-    }
-
     if ([self.delegate respondsToSelector:@selector(playbackServiceDidSwitchAspectRatio:)]) {
         [_delegate playbackServiceDidSwitchAspectRatio:_currentAspectRatio];
     }
 }
 
-- (NSString *)stringForAspectRatio:(VLCAspectRatio)ratio
-{
-    switch (ratio) {
-            case VLCAspectRatioFillToScreen:
-            return NSLocalizedString(@"FILL_TO_SCREEN", nil);
-            case VLCAspectRatioDefault:
-            return NSLocalizedString(@"DEFAULT", nil);
-            case VLCAspectRatioFourToThree:
-            return @"4:3";
-            case VLCAspectRatioSixteenToTen:
-            return @"16:10";
-            case VLCAspectRatioSixteenToNine:
-            return @"16:9";
-        default:
-            NSAssert(NO, @"this shouldn't happen");
-    }
-}
-
 - (void)setVideoTrackEnabled:(BOOL)enabled
 {
     if (!enabled)
@@ -1407,23 +1439,39 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
 #if TARGET_OS_IOS
 - (void)_recoverLastPlaybackState
 {
-    VLCMLMedia *media = [VLCMLMedia mediaForPlayingMedia:_mediaPlayer.media];
-    if (!media) return;
+    VLCMedia *media = _mediaPlayer.media;
+    VLCMLMedia *libraryMedia = [VLCMLMedia mediaForPlayingMedia:media];
+    if (!libraryMedia) return;
 
-    CGFloat lastPosition = media.progress;
-    // .95 prevents the controller from opening and closing immediatly when restoring state
-    //  Additionaly, check if the media is more than 10 sec
-    if (lastPosition < .95
-        && media.duration > 10000
-        && _mediaPlayer.position < lastPosition) {
+    VLCMediaList *mediaList = [_listPlayer mediaList];
+    NSUInteger mediaIndex = [mediaList indexOfMedia:media];
+
+    /* continue playback for the first item even if repeating */
+    if (self.repeatMode != VLCDoNotRepeat && mediaIndex != _itemInMediaListToBePlayedFirst) {
+        goto bailout;
+    }
+
+    CGFloat lastPosition = libraryMedia.progress;
+    SInt64 mediaDuration = libraryMedia.duration;
+
+    if (_mediaPlayer.position < lastPosition) {
         NSInteger continuePlayback;
-        if (media.type == VLCMLMediaTypeAudio)
+        if (libraryMedia.type == VLCMLMediaTypeAudio) {
+            if (!libraryMedia.isPodcast) {
+                goto bailout;
+            }
             continuePlayback = [[[NSUserDefaults standardUserDefaults] objectForKey:kVLCSettingContinueAudioPlayback] integerValue];
-        else
+        } else {
+            if (mediaDuration < 10000 && !libraryMedia.isExternalMedia) {
+                goto bailout;
+            }
             continuePlayback = [[[NSUserDefaults standardUserDefaults] objectForKey:kVLCSettingContinuePlayback] integerValue];
+        }
 
         if (continuePlayback == 1) {
-            [self setPlaybackPosition:lastPosition];
+            if (lastPosition * mediaDuration > 120000.) {
+                [self setPlaybackPosition:lastPosition];
+            }
         } else if (continuePlayback == 0) {
             NSArray<VLCAlertButton *> *buttonsAction = @[[[VLCAlertButton alloc] initWithTitle: NSLocalizedString(@"BUTTON_CANCEL", nil)
                                                                                          style: UIAlertActionStyleCancel
@@ -1436,12 +1484,14 @@ NSString *const VLCPlaybackServicePlaybackDidMoveOnToNextItem = @"VLCPlaybackSer
             UIViewController *presentingVC = [UIApplication sharedApplication].delegate.window.rootViewController;
             presentingVC = presentingVC.presentedViewController ?: presentingVC;
             [VLCAlertViewController alertViewManagerWithTitle:NSLocalizedString(@"CONTINUE_PLAYBACK", nil)
-                                                 errorMessage:[NSString stringWithFormat:NSLocalizedString(@"CONTINUE_PLAYBACK_LONG", nil), media.title]
+                                                 errorMessage:[NSString stringWithFormat:NSLocalizedString(@"CONTINUE_PLAYBACK_LONG", nil), libraryMedia.title]
                                                viewController:presentingVC
                                                 buttonsAction:buttonsAction];
 
         }
     }
+
+    bailout:
     [self restoreAudioAndSubtitleTrack];
 }
 
